/*
 * Copyright (c) 2018 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-Nordic-5-Clause
 */

/** @file
 *  @brief Nordic UART Bridge Service (NUS) sampleinclude(${ZEPHYR_BASE}/boards/common/nrfjprog.board.cmake)
 */


#include <stdio.h>
#include <stddef.h>
#include <zephyr/drivers/sensor.h>
#include <zephyr/types.h>
#include <zephyr/kernel.h> 
#include <zephyr/device.h>
#include <soc.h>


#include <zephyr/bluetooth/bluetooth.h>
#include <zephyr/bluetooth/uuid.h>
#include <zephyr/bluetooth/gatt.h>
#include <zephyr/bluetooth/hci.h>

#include <bluetooth/services/nus.h>
//#include <bluetooth/scan.h>

#include <zephyr/drivers/gpio.h>

#include <zephyr/settings/settings.h>
#include <zephyr/pm/pm.h>
#include <zephyr/pm/device.h>
#include <zephyr/pm/policy.h>
#include <zephyr/sys/reboot.h>
#include <stdio.h>


#include "common.h"


#define STACKSIZE 4096
#define PRIORITY 7

uint8_t FWVERSION[]="1.9.9";    


 K_FIFO_DEFINE(fifo_cmd_data);                          // the fifo to proc_cmd()
 K_FIFO_DEFINE(fifo_ble_data);                          // the fifo to ble_write()
struct k_fifo fifo_uart_tx_data;                        // the fifo to the uart (the urt commnicates with the Quectel module)
struct k_fifo fifo_uart_rx_data;                        // the fifo from the uart (the Quectel module)
struct k_fifo fifo_bt_data;                             // the inermedite ble_write data fifo
                                                        // bt_send_data gets a packet, put's it into the fifo
                                                        // bt-send-data_proper is its own task that parses the data, then sends it on to ble_write (another task) via fifo_ble_data

K_SEM_DEFINE(ble_init_ok,6,6);                          // signals to virutally all the tasks that lvbuetoothis running - allowing them to start
K_SEM_DEFINE(irq_active,0,6);                           // this seamphore is local to the ST-Micro RFAL stuff. Thr RFALpackage has an interrupt handler that is imcompatable wuth
K_SEM_DEFINE(irq_done,0,6);                             // with Zephyr - I put in a work-around using these two semaphors
K_SEM_DEFINE(run_script,0,6);                           // specific to the SCRIPT command                     
bool PROCRUNNING=false;                                 // set when proc_cmd() is running;
bool SUPERUSERMODE=false;
bool RUN_SPOOF=false;                                   // this is set by lock_cmd. At the end of procPcmd - it the switch is set to true the spoof timer is started
bool RUN_CHECK_SWS=false;                               // this is set by lock_cmd. At the end of procPcmd - it the switch is set to true the check_sws timer is started
struct PARAMS PARMS;
uint32_t KEYPAD_ERR_CT;                                 // number of successive keypad errors we've encountefred
uint32_t KEYPAD_REACTIVATE=0;                           // the keypad unlock tine - set when KEYPAD_ERR_CT is reached. 
uint32_t KEYPAD_IDLE_LIMIT;                             // the keypad idle timeut 
int32_t FLASHER_MODE;                                   // 1 IS FOR QUECTEL - FLSH RED/GREEN AT 1/2 SECIND INTERVALS
                                                        // 2  IS FOR KEYPAD LOCKED OUT - FLASH RED ON AND OFF WHILE KEYPAD IS LOCKED OUT
                                                        // 3  IS FOR THE SIXTY SECND DELAY BEFORE UPDATING AWS AFTER A DOOR OR HASP CLOSE
int32_t FLASHER_LIMIT;                                  // FOR MODE 2, THE NUMBER OF TIMES TO FLASH (PARMS.ERRTMO*2)
int32_t FLASHER_CONTROL;                                // FOR MODE 3, I WANT THE GREEN LED TO FLASH FOR 200 MILLISECONDS EVERY FIVE SECONDS
uint32_t TIMEOUT;                                       // set by lockcmd.c, it is used to set the open_alarm and tag_alarmtimers

uint32_t TIMEOUTDELAY;                                  // this delay is how long the timeout for door open and tag present - it's segt to two minutes - change for production
bool KEYPAD_TIMED_OUT;
bool RTCACTIVE=false;
uint16_t VTBL[3000];                                   // voltages tabled by the voltage monitoring task
volatile bool NO_BT_DISP;
bool TIMER_RUNNING=false;
bool DOOR_OPEN=false;
bool TAG_MISSING=false;
bool VDD2_UP=false;
extern struct k_sem voltage_mon_on;
extern struct k_sem voltage_mon_off;
extern volatile bool RFPAUSE;
extern volatile bool RFRUN;
extern uint16_t LOW_VOLTAGE;                                                           // the lowest battery voltage seen while the motor is spinning
extern uint32_t LVCNT;
extern volatile uint8_t UART_RECD_DATA[300];
static struct bt_conn *current_conn;
static struct bt_conn *auth_conn;
const struct device *gpio0; 
const struct device *gpio1;
const struct device *spi0;
const struct device *i2c1;
const struct device *uart0;
const struct device *i2s0;  
const struct device *adc0;
struct gpio_callback keypad_cb_data_gpio0; 
struct gpio_callback keypad_cb_data_gpio1;
struct k_timer bootclock;                                                   // controls rhe autmatic reboot on Sunday morming
struct k_timer msclock;                                                     // this is a gneral use milllisecomd timer
struct k_timer flasher;                                                     // for flashing red and green leds 
struct k_timer spoof;                                                       // triggers the door/hasp closed logic when the door/hasp was never opened
struct k_timer update_delay;                                                // delays quectel update after the door or hasp has closedd
struct k_timer alarm;                                                       // an alarmclock that is primarily used in the Quectel logic
struct k_timer check_sws;                                                   // usd to check the door (or hasp) switch - waiting for it to close
struct k_timer open_alarm;                                                  // used to check if the door has been opened for a lomng long time
struct k_timer tag_alarm;                                                   // used to check the presence/absence of a rfid tag
bool DUMMYREAD;                                                             // This is dhared between vdd2_up() and eeprom_retrieve
                                                                            //  When i2C is RESUMED the EEPROM read fails
                                                                            // vdd2_up sets this when t does the first read
uint32_t  DOOR_OPENED_TIME;                                                 // used for the RFID TAG TEST in DOOR_CLOSE. IF less than two minutes have passed we will
                                                                            // lock the door and start a timer 
                                                                         
uint32_t  DOOR_PHASE;                                                       // used by the DOOR CLOSE routine to figure out what's going on
uint32_t TAG_NOT_PRESENT_TIME;                                              // the time the door was closed without a tag (two minute rule)
bool WRONG_TAG;                                                                                                                                         
bool DOORCLOSED;
bool BOOTING=true;
bool KEYPADACTIVE=false;
bool DOORACTIVE=false; 
bool HASPACTIVE=false; 
bool UPDATEPENDING=false;
bool TESTING=false;
bool BTON=false;
bool RFALFTA=true;
bool RFALALARM=false;
bool QPOWEREDUP=false;
bool QCONNECTED=false;
int32_t HOURSTOREBOOT;
volatile bool RED_ON=false;
volatile bool GREEN_ON=false;
bool LOGGING=false;
bool RUNNINGSCRIPT=false;
uint8_t SCRIPTCMD1[10];
uint8_t SCRIPTCMD2[10];
int32_t SCRIPTCNT;
bool SCRIPTPAUSE;
bool SCRIPTCMDDONE;
uint8_t PATTERN[20];
uint8_t PATTERNREC[1100];
uint8_t PREVREC[100];
uint8_t RTCRESETCODE;
bool ALARM;
uint8_t RFIDTAGID[30];
          
struct k_mutex rfalirqlock;      
struct k_mutex rfalcommlock;    
struct k_mutex rfalworkerlock;    
struct k_mutex quelock;   
struct k_sem irq_active;
struct k_sem irq_done;
                                           
uint32_t MILLITIME;
struct EEPROMMAP MAP;
struct EEPROMMAP MAP2;
uint16_t IDLEVOLTAGE;
extern uint16_t LOW_VOLTAGE;
k_tid_t CURRENT_THREAD;
uint8_t bt_name[11];
uint8_t CMD[CMDSIZE];
uint8_t LASTCMD[CMDSIZE];                                                           // the last command that was successfully run
uint32_t LASTCMDDT;                                                                 // the (POSIX) time it executed
bool CONNECTED=false;
bool WAITING_FOR_MECH=false;
static const struct bt_data ad[] = {
	BT_DATA_BYTES(BT_DATA_FLAGS, (BT_LE_AD_GENERAL | BT_LE_AD_NO_BREDR)),
        BT_DATA(BT_DATA_NAME_COMPLETE,bt_name, 10),

};
uint32_t BT_ADVERTISMENT_MIN, BT_ADVERTISMENT_MAX;                    // used to set advertisment rate
#define BT_LE_ADV_CONN_SLOW BT_LE_ADV_PARAM(BT_LE_ADV_OPT_CONNECTABLE, \
				       BT_ADVERTISMENT_MIN, \
				      BT_ADVERTISMENT_MAX, NULL)
static const struct bt_data sd[] = {
	BT_DATA_BYTES(BT_DATA_UUID128_ALL, BT_UUID_NUS_VAL),
};


/* Prevent deep sleep (system off) from being entered on long timeouts
 * or `K_FOREVER` due to the default residency policy.
 *
 * This has to be done before anything tries to sleep, which means
 * before the threading system starts up between PRE_KERNEL_2 and
 * POST_KERNEL.  Do it at the start of PRE_KERNEL_2.
 */
static int disable_ds_1(const struct device *dev)
{
	ARG_UNUSED(dev);

	pm_policy_state_lock_get(PM_STATE_SOFT_OFF, PM_ALL_SUBSTATES);
	return 0;
}

SYS_INIT(disable_ds_1, PRE_KERNEL_2, 0);
void spoof_tick(struct ktimer *clk)
{
    if(PARMS.LOCKSTATE=='D')
         DOORACTIVE=true;
    else HASPACTIVE=true;
    k_wakeup(CURRENT_THREAD); 
}

void update_delay_tick(struct ktimer *clk)
{
    UPDATEPENDING=true;
    k_wakeup(CURRENT_THREAD); 
}
void flasher_tick(struct ktimer *clk)
{
   uint32_t i;
   if(FLASHER_MODE==1)                                      // alterate fashing the RED and GREEN LEDs
      if(!RED_ON)
      {
        RED_ON=true;
        GREEN_ON=false;
        gpio_pin_set(gpio0,LED_RED,1);
        gpio_pin_set(gpio0,LED_GREEN,0);
        
      }
   else {
            RED_ON=false;
            GREEN_ON=true;
            gpio_pin_set(gpio0,LED_RED,0);
            gpio_pin_set(gpio0,LED_GREEN,1);
         }
    if(FLASHER_MODE==2)                                        // flash the RED LED on and off 
    {   gpio_pin_set(gpio1,VDD2_ENABLE,1);
        if((FLASHER_LIMIT%4)!=0)
           {
            gpio_pin_set(gpio0,LED_RED,0);
            RED_ON=false;
           }
       else
             {
               gpio_pin_set(gpio0,LED_RED,1);
               RED_ON=true;
             }
        FLASHER_LIMIT--;
        if(FLASHER_LIMIT<1)
          k_wakeup(CURRENT_THREAD);
    }
    if(FLASHER_MODE==3)                                               // Flash the Green Led on for 200 milliseconds every two seconds
    {
        i=FLASHER_CONTROL%20;
        FLASHER_CONTROL++;
        if(i==0)                                                      // if the green LED isn't on
             {
             if(!VDD2_UP)                                             // if VDD2 isn't up
               gpio_pin_set(gpio1,VDD2_ENABLE,1);                     //      bring up the LDO
             gpio_pin_set(gpio0,LED_GREEN,1);                         // turn on the green LED                                        // and set the logical to show that it's lit
             }
        if(i==1)
            {
                 if(!VDD2_UP)                                         // if the green led is on and the VDD2_UP route isn't active
                      gpio_pin_set(gpio1, VDD2_ENABLE,0);             // bring down the LDO
                 gpio_pin_set(gpio0,LED_GREEN,0);                     // turn off the led logical
                 GREEN_ON=false;
             }
    }

}
void boottick(struct ktimer *clk)
{
  HOURSTOREBOOT--;                                                 // DECREMENT THE HOURS TO REBOOT
  RTCACTIVE=true;                                                  // set the RTCACTIVE flag
  k_wakeup(CURRENT_THREAD);                                        // and go wake up the current threed (cycle through main)
}

void check_open_tick(struct ktimer *clk)
{
    DOOR_OPEN=true;                                                // set the DOOR_OPEN flag
    k_wakeup(CURRENT_THREAD);                                      // and go wake up the current threed (cycle through main)
    
}
void check_tag_tick(struct ktimer *clk)
{
    TAG_MISSING=true;                                              // set the TAG_IS_MISSING flag
    k_wakeup(CURRENT_THREAD);                                      // and go wake up the current threed (cycle through main)
    
}
void check_sws_tick(struct ktimer *clk)                            // the delay loop that runs after the door or hasp has been opened
{
    uint32_t ret;
    if(!VDD2_UP)
          gpio_pin_set(gpio1,VDD2_ENABLE,1);
    //k_sleep(K_MSEC(1000))
    for(ret=0;ret<10000;ret++);                                     // very short delay to allow the HALL EFFECT SENSORS to come up
    if(PARMS.LOCKSTATE=='D')                                       // if the door is open
    {
       ret=gpio_pin_get(gpio0,DOORSW);                             // read the state of the door Hall Effect Sensor
       if(ret==0)                                                  // if the sensor is  0 (the door is open)
       {
          if(!VDD2_UP)
               gpio_pin_set(gpio1,VDD2_ENABLE,0);
           return;                                                 // and we're done till the next tick         
       }
       DOORACTIVE=true;                                            // the door has been closed - set the has active switch
       //gpio_pin_set(gpio1,VDD2_ENABLE,0);
       k_wakeup(CURRENT_THREAD);                                   // and go wake up the current threed (cycle through main)
       return;
    }
    if(PARMS.LOCKSTATE=='H')                                       // if the HASP is open
    {
       ret=gpio_pin_get(gpio0,HASPSW);                             // read the state of the hasp Hall Effect Sensor
       if(ret==0)                                                  // if the sensor is  0 (the hasp is open)
       {
            if(!VDD2_UP)
               gpio_pin_set(gpio1,VDD2_ENABLE,0);
           return;                                                 // and we're done till the next tick          
       }
   
       HASPACTIVE=true;                                            // the hasp has been closed - set the has active switch
       //gpio_pin_set(gpio1,VDD2_ENABLE,0);
       k_wakeup(CURRENT_THREAD);                                   // and go wake up the current threed (cycle through main)
       return;
    }
    if(!VDD2_UP)
        gpio_pin_set(gpio1,VDD2_ENABLE,0);                             // if the door (or hasp) is still open bring down VDD2

}  
void mstick(struct ktimer *clk)
{MILLITIME++;
} 
void alarmrtn(struct ktimer *clk)
{
 ALARM=true;
}
void init_timers()                                                                  // initialize the timers
{
k_timer_init(&msclock,mstick, NULL);
k_timer_init(&bootclock,boottick, NULL);
k_timer_init(&alarm,alarmrtn, NULL);
k_timer_init(&flasher,flasher_tick, NULL);
k_timer_init(&spoof,spoof_tick, NULL);
k_timer_init(&update_delay,update_delay_tick, NULL);
k_timer_init(&check_sws,check_sws_tick, NULL);
k_timer_init(&open_alarm,check_open_tick, NULL);
k_timer_init(&tag_alarm,check_tag_tick, NULL);

} 
static void exchange_func(struct bt_conn *conn, uint8_t err, struct bt_gatt_exchange_params *params)
{
	//if (!err) {
	//	printk("MTU exchange done");
	//} else {
	//	printk("MTU exchange failed (err %" PRIu8 ")", err);
	//}
}

static void connected(struct bt_conn *conn, uint8_t err)
{
	char addr[BT_ADDR_LE_STR_LEN];
    uint8_t buff[50];
    CONNECTED=true;
    SUPERUSERMODE=false;                                             // turn off the superuser switch
	if (err) {
		if(!LOGGING)
           logmsg_add(2,"Connection failed ");
		return;
	}
	static struct bt_gatt_exchange_params exchange_params;

	exchange_params.func = exchange_func;
	err = bt_gatt_exchange_mtu(conn, &exchange_params);
	//if (err) {
	//	printk("MTU exchange failed (err %d)", err);
	//}

   sprintf(buff,"MTU size is: %d\n", bt_gatt_get_mtu(conn));
    err = bt_gatt_exchange_mtu(conn, &exchange_params);
    sprintf(buff,"MTU size is: %d\n", bt_gatt_get_mtu(conn));
	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));
  
	current_conn = bt_conn_ref(conn);
    vdd2_up();


}

static void disconnected(struct bt_conn *conn, uint8_t reason)
{
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));
    CONNECTED=false;
	//LOG_INF("Disconnected: %s (reason %u)", log_strdup(addr), reason);

	if (auth_conn) {
		bt_conn_unref(auth_conn);
		auth_conn = NULL;
	}

	if (current_conn) {
		bt_conn_unref(current_conn);
		current_conn = NULL;
	
	}
    //TESTING=false;                                         // turn off the testing switch
    SUPERUSERMODE=false;                                     // turn off the superuser switch
    if(!WAITING_FOR_MECH)                                    // if we're not waiting for soething to happen with the door or hasp                                            
         vdd2_down();                                        // bring down vdd2             

}




BT_CONN_CB_DEFINE(conn_callbacks) = {
	.connected    = connected,
	.disconnected = disconnected,

};
void bt_send_data(uint8_t *data)
{
 uint32_t i;
 static struct bt_data_t btbuff,*buf;
 if(!CONNECTED)                                                                     // sometimes we get here when we're not connected
      return;       
 
 i=strlen(data);
 buf=k_malloc(sizeof(btbuff));                                                       // create a  fifo record
 memcpy(buf->data,data,i);
 if(buf->data[i-1]!=RS)
 {
    buf->data[i]=RS;
    i++;
 }
 
 buf->len=i;                                                                        //      and the length  
 k_fifo_put(&fifo_ble_data,buf);                                                    // and push it into the fifo  

}

void bt_send_data_proper()
{
 uint8_t btdata[200];
 uint32_t btlen;
 uint32_t xmitlen;
 volatile uint32_t maxbuff;
 uint32_t i,j;
 static struct ble_data_t btbuff,*buf;
 k_sem_take(&ble_init_ok, K_FOREVER);                                               // wait for the program to initialize
 for(;;)                                                                            // loop forrever
 {
    buf = k_fifo_get(&fifo_bt_data, K_FOREVER);                                     // wait for a packet from bt_send_data
    memset(btdata,0,sizeof(btdata));                                                // clear the buffer to nulls
    maxbuff=20;                                                                     // set the BLE MTU
    memcpy(btdata,buf->data,buf->len);                                              // retrieve the data
    k_free(buf);                                                                    // we're done with the fifo enty - get rid of it
    btlen=strlen(btdata);                                                           // see how long it is
    if(btdata[btlen-1]!=RS)                                                         // if there isn't a record seperator at the end of the string
    {
       btdata[btlen]=RS;                                                            //     append the record seperator
       btlen++;
    }
     for(;;)
        {
         if(btlen>maxbuff)                                                              // we have to senbd it in UART_BUF_LEN chunks
        xmitlen=maxbuff;                                                            //     so if it's longer than ble data
        else xmitlen=btlen;                                                            // we iit hiw much we're goin to send
        buf=k_malloc(sizeof(btbuff));                                                 // create a  fifo record
        memcpy(buf->data,&btdata,xmitlen);                                             // copy in the data
        buf->len=xmitlen;                                                              //      and the length  
        k_fifo_put(&fifo_ble_data,buf);                                                // and send it off to be written           
        if(btlen<=maxbuff)                                                              // if we've written everything
            break;                                                                      // get out
        j=0;                                                                            // otherwise
        for(i=maxbuff;i<sizeof(btdata);i++)                                             // shuffle out the data we just sent
        {
            btdata[j]=btdata[i];
           j++;
        }
        btlen=strlen(btdata);                                                           // and resend
        k_sleep(K_MSEC(200));                                                           // after a brief delay
     
        }
 }
}
K_THREAD_DEFINE(bt_send_data_proper_thread_id, STACKSIZE, bt_send_data_proper, NULL, NULL,
		NULL, PRIORITY, 0, 0);
static void bt_receive_cb(struct bt_conn *conn, const uint8_t *const data,
			  uint16_t len)
{
	static int i,j,msgcnt=0;
    static bool foundrs=false;
    static uint8_t *cmdptr=NULL,cmdbuff[200];
	static char addr[BT_ADDR_LE_STR_LEN] = {0};
    static uint8_t xdata[100],*ptr1,*ptr2; 
    static struct cmd_data_t *buf,fifo_buff;
    static uint8_t cmd[100];
    static uint32_t cmdcnt=0;
    if(PROCRUNNING)                                           // if proc_cmd() is running
       return;                                                // ignore the traffic

    msgcnt++;
	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, ARRAY_SIZE(addr));

	//LOG_INF("Received data from: %s", log_strdup(addr));
        if(cmdptr==NULL)                                      // if the command pointer is NULL
            {cmdptr=cmdbuff;                                  // point to the begiinning of the ommand buffer
             memset(cmdbuff,0,sizeof(cmdbuff));               // and clear the command buffer to NULLS

            }        
        memset(xdata,0,sizeof(xdata));
        if(len<100)                                           // retrieve the data from the blue tooth buffer
            memmove(xdata,data,len);
        else memmove(xdata,data,99);
        i=strlen(xdata);                                      // find the end of the bluetooot data
        //if(xdata[i-1]=='\r')                                  // if the bluetooth data terminated with a 
          // xdata[i-1]=0;                                       //     return - get rid of it
        strcat(cmdbuff,xdata);                                // now add the data just received to the command buffer
        i=strlen(cmdbuff);  
        cmdptr=&cmdbuff[i];                                  // move the command pointter to the end of 
                                                             // what's already been stored
        if(cmdbuff[i-1]!=RS)                                  // if the command didn't end with a RS
             return;                                          //      we're done;
        buf=k_malloc(sizeof(fifo_buff));

        memset(buf->data,0,sizeof(buf->data));                        // otherwise, build the entry to pass to
        memcpy(buf->data,cmdbuff,strlen(cmdbuff));  
        i=strlen(cmdbuff);                                    // proccmd
        buf->len=i;
        k_fifo_put(&fifo_cmd_data,buf);                     // and go process the command
        cmdptr=NULL;
 

}


static struct bt_nus_cb nus_cb = {
	.received = bt_receive_cb,
};

void error(void)
{


	while (true) {
		/* Spin for ever */
		k_sleep(K_MSEC(5000));
	}
}
void vdd2_up()
{int32_t err, dummy;
 uint8_t buff[50], buffer[50];
 uint16_t voltage;
 //if(!LOGGING)                                                              // if we're not booting up
  //    err = bt_le_adv_stop();                                              //      stop advertising
 //k_sleep(K_MSEC(2000));
 if(VDD2_UP)                                                               // if VDD2 is already up
        return;                                                            // get out
 gpio_pin_set(gpio1,VDD2_ENABLE,1);                                        // powerup VDD2
 gpio_pin_set(gpio0,LED_ILLUM,1);                                          // and the illumintion LED
 err = pm_device_action_run(i2c1, PM_DEVICE_ACTION_RESUME);                // restart I2C
 memset(&MAP2,0,sizeof(MAP2));
 //DUMMYREAD=true;
 //eeprom_retrieve(0,sizeof(MAP2),&MAP2);                                    // do  dummy read on the EEPROM
 //DUMMYREAD=false;
 

 PARMS.LASTBATT=voltage;                                                   // save it in the oarams record
 if ((!KEYPADACTIVE)&&(!LOGGING)&&(!BOOTING))                              // if thus isn't from the KEYPAD and  we're not booting up AND WE ARE LOGGGING
    eeprom_update_parms();                                                 //     update the EEPROM 

VDD2_UP=true;         





}
void vdd2_down()
{int32_t err;
 uint16_t voltage1, voltage2;
 uint8_t buff[50];

 err = pm_device_action_run(i2c1, PM_DEVICE_ACTION_SUSPEND);                 // stop I2C

 if(QPOWEREDUP)                                                              // if the quectel chip is powered up
     quectel_powerdown();                                                    // go power it down (thus suspends uart0)
 gpio_pin_set(gpio0,LED_ILLUM,0);                                            // turn off the keypad LEDs
 gpio_pin_set(gpio1,VDD2_ENABLE,0);                                          // power down VDD2
 if(!CONNECTED)                                                              // if we're not connected
 {                                                                           // we're going to set idlevoltge

       k_sleep(K_MSEC(100));                                                 // waiting 100 msec between reade
 
       if(voltage2>voltage1)
           voltage1=voltage2;
       IDLEVOLTAGE=voltage1;                                                 // we save the higher  of the two voltages
 }
 VDD2_UP=false;                                                              // mark VDD2 down 
                 
}
 


void error_flash(uint32_t first,uint32_t second, uint32_t third)
{

      
      if(first==0)  
      {

      
             gpio_pin_set(gpio0,LED_RED,1);                           // if it isn't, turn on the RED led
             k_sleep(K_MSEC(1000));                                   // sleep for two seconds
             gpio_pin_set(gpio0,LED_RED,0);                           // and turn it off
             k_sleep(K_MSEC(1000));
      }
      else {
             gpio_pin_set(gpio0,LED_GREEN,1);                         // if it isn't, turn on the RED led
             k_sleep(K_MSEC(1000));                                   // sleep for two seconds
             gpio_pin_set(gpio0,LED_GREEN,0);                          // and turn it off
             k_sleep(K_MSEC(1000));

           }
      if(second==0)  
      {

      
             gpio_pin_set(gpio0,LED_RED,1);                           // if it isn't, turn on the RED led
             k_sleep(K_MSEC(1000));                                   // sleep for two seconds
             gpio_pin_set(gpio0,LED_RED,0);                           // and turn it off
             k_sleep(K_MSEC(1000));
      }
      else {
             gpio_pin_set(gpio0,LED_GREEN,1);                         // if it isn't, turn on the RED led
             k_sleep(K_MSEC(1000));                                   // sleep for two seconds
             gpio_pin_set(gpio0,LED_GREEN,0);                          // and turn it off
             k_sleep(K_MSEC(1000));

           }
                 if(first==0)  
      {

      
             gpio_pin_set(gpio0,LED_RED,1);                           // if it isn't, turn on the RED led
             k_sleep(K_MSEC(1000));                                   // sleep for two seconds
             gpio_pin_set(gpio0,LED_RED,0);                           // and turn it off
             k_sleep(K_MSEC(1000));
      }
      else {
             gpio_pin_set(gpio0,LED_GREEN,1);                         // if it isn't, turn on the RED led
             k_sleep(K_MSEC(1000));                                   // sleep for two seconds
             gpio_pin_set(gpio0,LED_GREEN,0);                          // and turn it off
             k_sleep(K_MSEC(1000));

           }
      if(third==0)  
      {

            gpio_pin_set(gpio1,VDD2_ENABLE,1);
             gpio_pin_set(gpio0,LED_RED,1);                           // if it isn't, turn on the RED led
             k_sleep(K_MSEC(1000));                                   // sleep for two seconds
             gpio_pin_set(gpio0,LED_RED,0);                           // and turn it off
             k_sleep(K_MSEC(1000));
      }
      else {
             gpio_pin_set(gpio0,LED_GREEN,1);                         // if it isn't, turn on the RED led
             k_sleep(K_MSEC(1000));                                   // sleep for two seconds
             gpio_pin_set(gpio0,LED_GREEN,0);                          // and turn it off
             k_sleep(K_MSEC(1000));

           }
         
         
         

}
int32_t advertising_start()
{
  int32_t err;
  bool vdd2;
  uint8_t buff[50];
  vdd2=VDD2_UP;
  if(!vdd2)
      vdd2_up();
  err = bt_le_adv_start(BT_LE_ADV_CONN_SLOW, ad, ARRAY_SIZE(ad), sd,  ARRAY_SIZE(sd));
  if (err) {
                sprintf(buff,"BT ADV START RETURNED %d",err);
                logmsg_add(2,buff); 
           }
  if(!vdd2)
      vdd2_down();
  return err;

}

void keypad_disable()
{
    int32_t ret;
    ret=gpio_remove_callback(gpio0, &keypad_cb_data_gpio0);                 // disable the keypad callback on gpio0        
    ret=gpio_remove_callback(gpio1, &keypad_cb_data_gpio1);                 // disable the keypad callback on gpio1
}

void advertising_stop()
{
    int32_t err;
    uint8_t buff[50];
    bool vdd2;
    vdd2=VDD2_UP;
    if(!vdd2)
         vdd2_up();
    err = bt_le_adv_stop();
    if (err) {
                  sprintf(buff,"BT ADV STOP RETURNED %d",err);
                  logmsg_add(2,buff); 
             }
    if(!vdd2)
       vdd2_down();
    return err;
}
void main(void)
 {	static int err = 0;
    uint32_t boot_reason;
    uint32_t millivolts;
    uint32_t test_time, curr_time, i,no_key_cnt;;
    bool tag_present;

    uint32_t tag_result;
    int32_t testbatt;
    struct ACTREC actrec;
    uint8_t buff[100];
    struct RTCDATE xdate;

    



        uint8_t buffer[30], *ptr, buffer2[]="DATA\r";
        bool uartok;
        CURRENT_THREAD=k_current_get();
        while(true)
             que_fw_download1();             
      
        init_gpio();
     
        init_i2c();
 

        init_timers();                          
        k_fifo_init(&fifo_uart_tx_data);
	    k_fifo_init(&fifo_uart_rx_data);
        k_fifo_init(&fifo_bt_data);
        err=init_uart();
        //k_sleep(K_MSEC(50000));  
        uartok=device_is_ready(uart0);
        if(uartok)
          strcpy(buffer,"UART UP");
        else strcpy(buffer, "UART DOWN");
        memset(LASTCMD,0,sizeof(LASTCMD));                              // clear the last command to NULLS
	    IDLEVOLTAGE=millivolts;                                         // we save this so we can put it in activity rcords

        vdd2_up();                                                      // bring up vdd2

     
        k_sleep(K_MSEC(10));                                            // brief pause for power up
        eeprom_retrieve(SIG_ADDRESS,SIG_SIZE,buffer);                   // retrieve the EEPROM signature
 
        err=strncmp(SIG_CONTENT,buffer,SIG_SIZE);                       // check to see if the SIGNATURE is in place
        if(err!=0)                                                      // if it isnt,
             eeprom_init();                                             //     reinitiaize the EEPROM
        eeprom_retrieve(0,sizeof(MAP),&MAP);                            // now -load the eeprom map
        eeprom_retrieve(MAP.addr[9],sizeof(PARMS),&PARMS);              // and the params
        TIMEOUTDELAY=PARMS.TIMEOUTDELAY;                                          // load the Alert timeout
        build_offsets();                                                // and go build the parameters offset table

        boot_reason=NRF_POWER->RESETREAS;                               // retrieve the bootr reason code


        NRF_POWER->RESETREAS=0XFFFFFFFF;                                // and clear the the register
        i=strcmp(FWVERSION,PARMS.FWVERSION);                            // see if there has been a firmware update
        if(i!=0)                                                        // if there has
        {
            sprintf(buffer,"UPDATE %s",PARMS.FWVERSION);                // log the change in messages
            if(!LOGGING)
               logmsg_add(2,buffer);
            sprintf(buffer,"    TO %s",FWVERSION);
            if(!LOGGING)
                logmsg_add(2,buffer);
            memset(PARMS.FWVERSION,0,sizeof(PARMS.FWVERSION));          // and update parms
            strcpy(PARMS.FWVERSION,FWVERSION);
            eeprom_update_parms();                                      // then rewwrtite them
 
        }
        testbatt=millivolts-PARMS.LASTBATT;                            // see if battery voltage has increased significantly
        if((boot_reason==0)&&(millivolts>6200)&&(testbatt>1000))       // if it has - and this is a power on boot -- and these btteries look new
        {
            for(i=0;i<6;i++)                                            // look ffor an empty battery history slot in the PARMS record
            {
                if(PARMS.batt[i].battdt==0)
                   break;
            }
            if(i==6)                                                    // if there wasn't an empty slot
            {
                for(i=1;i<6;i++)                                        // shuffle the oldest one off
                    PARMS.batt[i-1]=PARMS.batt[i];
                i=5;                                                    // point to the newest entry
            }
            PARMS.batt[i].battdt=dt_to_posix(get_rtc());                // fill out the battery history
            PARMS.batt[i].battbef=PARMS.LASTBATT;
            PARMS.batt[i].battaft=millivolts;
            if(!LOGGING)
                 logmsg_add(2,"NEW BATTERIES");
        }
        PARMS.LASTBATT=millivolts;                                       // in any case (whether we updated hist or not), update the lLASTBATT
        eeprom_store(MAP.addr[9],sizeof(PARMS),&PARMS);                  // and put the PARAMS back on the EEPROM
        
        while(true)                                                      // make sure the motor is centered
        {
            i=gpio_pin_get(gpio0,CNTRSW);                                // read the motor centered sensor
            if(i>0)                                                      // if it is
            { 
                if( PARMS.LOCKSTATE!='C')                                // make sure the proper lock state is recorded
                {                                                        // if it isn't
                    PARMS.LOCKSTATE='C';                                 //     set the lockstate to CENTERED
                    eeprom_update_parms();                               //     update the parameters onthe EEPROM
                    
                }
                break;                                                   //     and get out - everythng is fine

            }
            if(PARMS.LOCKSTATE=='X')                                     // if this is a bare board (LOCKSTATE  was initiaized but never set)
                 break;                                                  //     get out
            center_motorx(1);                                            //  the motor is not centered - center the                                

        }                                                                // and go retest the sensor
        /*                                                 
        while(true)
        {
            runrfid(0);
            runrfid(1);
            
        }                                               
        */
        vdd2_down();                                                   //   turn off VDD2 
    
        BT_ADVERTISMENT_MIN=PARMS.ADVINT*1000/625;                       // compute the advertising intervals
        BT_ADVERTISMENT_MAX=PARMS.ADVINT*1100/625;   
        memset(bt_name,0,sizeof(bt_name));                               // set up the device name
        memmove(bt_name,PARMS.lockid,10);     

      

       
	err = bt_enable(NULL);
	if (err) {
		error();
	}


 



	if (IS_ENABLED(CONFIG_SETTINGS)) {
		settings_load();
	}
    
	err = bt_nus_init(&nus_cb);
	if (err) {
		
		return;
	}

     err = advertising_start();
     setup_keyscan();
	if (err) {
		//LOG_ERR("Advertising failed to start (err %d)", err);
		return;
	}
  


    k_mutex_init(&quelock);                                                  // initialize all of the mutexes
    k_mutex_init(&rfalirqlock);
    k_mutex_init(&rfalcommlock);
    k_mutex_init(&rfalworkerlock);
    k_sem_give(&ble_init_ok);
    
    vdd2_up();
    BOOTING=true;                                                              // sset a logical to let get_rtc that wee're booting
    sprintf(buffer,"BOOT %d",boot_reason);                                     // We actually record the rboot here - since we may need to reset the RTC

    logmsg_add(0,buffer);                                                      // and log the boot in EEPROM here 

    if(RTCRESETCODE==99)                                                       // if there was a succesful RTC recovery (via AWS)
    {
        logmsg_add(2,"RTC RECOVEREVED FROM AWS");                              // log that info
        sys_reboot(SYS_REBOOT_COLD);                                           // and reboot
    }
    else if(RTCRESETCODE!=0)                                                   // if ther qwas a faiked attempt to recover the RTC
        {
            sprintf(buffer,"RTC REVOVERY FAILURE %d",RTCRESETCODE);            // build a log essage
            logmsg_add(2,buffer);                                              // ans recofrd it
            sys_reboot(SYS_REBOOT_COLD);                                       // and re-boot
        }
    PARMS.LASTBOOT=dt_to_posix(get_rtc());                                     // update the BOOT date in the PARAMS record
    eeprom_update_parms();                                                     // and go update the eeprom copy of params
    BOOTING=false;                                                             // turn off the BOOTING logical
    calc_reboot();                                                             // calculate hours to  reboot
    k_timer_start(&bootclock, K_MSEC(3600000), K_MSEC(3600000));                // start the reboot timer
    vdd2_down();
    setup_keyscan();            
  
	for (;;)                                                                   // and idle forever 
    {   k_sleep(K_FOREVER);
        if(DOOR_OPEN)                                                          // if the door(or hasp) has been open too long
            {
                k_timer_stop(&open_alarm);                                     // stop the timer
                DOOR_OPEN=false;                                               // reset the flag
                bt_send_data("OPEN DOOR ALARM");
                if(VDD2_UP)                                                    // if vdd2 is up - something is going on
                    {
  
                        TIMEOUT=TIMEOUT+10800;                                // add three hours to TIMEOUT
                        i=TIMEOUT-dt_to_posix(get_rtc());                       // time in seonds until we sound the door alarm
                        i=i*1000;
                        k_timer_start(&open_alarm,K_MSEC(i),K_NO_WAIT);       // set the timer to go off in TIMEOUTDELAY SECONDS
                        xdate=dt_to_disp(TIMEOUT);                            // testing only - 
                        sprintf(buff,"XXXXXXX %02d:%02d:%02d %02d/%02d/%04d",
                                 xdate.hour,xdate.minute,xdate.second,xdate.month,xdate.day,xdate.year);
                        bt_send_data(buff);                             
                        continue;                                             // and loop
                    }
                k_timer_stop(&check_sws);                                      // stop the door check timer
                vdd2_up();                                                     // bring up vdd2
                for(i=0;i<5;i++)
                {
   
                    k_sleep(K_MSEC(1500));   
                }
                
                if(!LOGGING)                                               
                    logmsg_add(2, "DOOR OPEN NOTIFICATION");                   // log the Alert
                if(PARMS.USEAWS=='Y')                                          // if we're using AWS
                {                                     
                    gpio_pin_set(gpio0,LED_ILLUM,0);                           //  turn off the ILLUMINATION LEDs
                    strcpy(CMD,"ALERTDOOR OR HASP IS OPEN,");
                    que_send_alert();
                }
                TIMEOUT=TIMEOUT+10800;                                          // Add three hours to TIMEOUT
                i=TIMEOUT-dt_to_posix(get_rtc());                               // time in seonds until we sound the door alarm
                i=i*1000;
                k_timer_start(&open_alarm,K_MSEC(i),K_NO_WAIT);                // set the timer to go off in two minutes
                xdate=dt_to_disp(TIMEOUT);                                     // testing only - 
                sprintf(buff,"NEW TIMEOUT %02d:%02d:%02d %02d/%02d/%04d",
                         xdate.hour,xdate.minute,xdate.second,xdate.month,xdate.day,xdate.year);
                bt_send_data(buff);      
                k_timer_start(&check_sws, K_MSEC(500), K_MSEC(500));            // and restart the check_sws timer
                vdd2_down();
                continue;                                                       // and go back to the main wait loop
            }
        if(TAG_MISSING)                                                          // if the door(or hasp) has been open too long
            {
                k_timer_stop(&tag_alarm);                                        // stop the timer
                TAG_MISSING=false;
                if(VDD2_UP)                                                      // if vdd2 is up - something is going on
                    {
                        bt_send_data("NO TAG ALARM");
                        TIMEOUT=TIMEOUT+10800;                                  // add three hours to TIMEOUT
                        i=TIMEOUT-dt_to_posix(get_rtc());                       // time in seonds until we sound the door alarm
                        i=i*1000;
                        k_timer_start(&tag_alarm,K_MSEC(i),K_NO_WAIT);       
                        xdate=dt_to_disp(TIMEOUT);                                   
                        sprintf(buff,"NEW TIMEOUT %02d:%02d:%02d %02d/%02d/%04d",
                                 xdate.hour,xdate.minute,xdate.second,xdate.month,xdate.day,xdate.year);
                        bt_send_data(buff);                             

                        continue;                                              // and loop
                    }
                vdd2_up();                                                     // bring up vdd2
                for(i=0;i<5;i++)
                {

                    k_sleep(K_MSEC(1500));   
                }
                
                if(!LOGGING)                                               
                    logmsg_add(2, "TAG MISSING NOTIFICATION");                          // log the Alert
                if(PARMS.USEAWS=='Y')                                          // if we're using AWS
                {                                     
                    gpio_pin_set(gpio0,LED_ILLUM,0);                           //  turn off the ILLUMINATION LEDs
                    strcpy(CMD,"ALERTNO TAG IN LOCK,");
                    que_send_alert();
                }

                TIMEOUT=TIMEOUT+10800;                                        // set the next timeout to three hours from now
                i=TIMEOUT-dt_to_posix(get_rtc());                              // time in seonds until we sound the tag alarm
                i=i*1000;                 
                k_timer_start(&tag_alarm,K_MSEC(i),K_NO_WAIT);                 // set the timer 
                xdate=dt_to_disp(TIMEOUT);                                     // testing only - 
                sprintf(buff,"NEW TIMEOUT %02d:%02d:%02d %02d/%02d/%04d",
                         xdate.hour,xdate.minute,xdate.second,xdate.month,xdate.day,xdate.year);
                bt_send_data(buff);   
                vdd2_down();

                continue;                                                       // and go back to the main wait loop
            }
        if(RTCACTIVE)                                                           // The reboot timer went off
            {
                RTCACTIVE=false;
                sprintf(buffer,"RTC HOURS LEFT %d",HOURSTOREBOOT);
                bt_send_data(buffer);
                if(HOURSTOREBOOT<0)                                             // if the reboot timer has expired
                    sys_reboot(SYS_REBOOT_COLD);                                //     reboot
                continue;
            }
        if(FLASHER_MODE!=0)
            {
                FLASHER_MODE=0;
                k_timer_stop(&flasher);
            }

		if(KEYPADACTIVE)                                                // if a key was pressed
            {
                if(KEYPAD_REACTIVATE>0)                                 // if the kaypad was locked out
                    {
                        test_time=KEYPAD_REACTIVATE;
                        vdd2_up();
                        curr_time=dt_to_posix(get_rtc());               // go get the currrent time
                        curr_time=dt_to_posix(get_rtc());               //    twice
                        if(curr_time>test_time)                         // if the lockout is over
                            {
                                KEYPAD_REACTIVATE=0;                    // reset the reactivate time
                                KEYPAD_ERR_CT=0;                        // and the error count
                                gpio_pin_set(gpio1,VDD2_ENABLE,1);      // bring up VDD2
                                gpio_pin_set(gpio0,LED_ILLUM,1);        // and the ILLUMINATION LEDs
                                getkeypadinput();                       // and go get the keypad input
                                KEYPADACTIVE=false;
                            }
                        else    {
                                    vdd2_down();
                                    KEYPADACTIVE=false;                     // otherise ignore the keypad input
                                }
                        }
                else    {
                            gpio_pin_set(gpio1,VDD2_ENABLE,1);          // bring up VDD2
                            gpio_pin_set(gpio0,LED_ILLUM,1);            // and the ILLUMINATION LEDs
                            getkeypadinput();                           // if there was no lockout, just proceed normLLY
                            KEYPADACTIVE=false;
                        }
            }    
        
        if(DOORACTIVE)                                          // the door switch toggled
            {

                k_timer_stop(&check_sws);                      // stop the check switch timer
                k_timer_stop(&spoof);                          // stop the spoof timer (it doesnt matter i it was neveer started)
                k_timer_stop(&open_alarm);                     // stop the door open timer
                k_timer_stop(&tag_alarm);                      // stop the tag alarm 
                DOOR_PHASE=0;                                  // set the DOOR PHASE to 0
                vdd2_up();                                     // bring up vdd2
                k_sleep(K_MSEC(1000));                         // wait for the RFID chip to stabilize
                struct LOGMSG msg;
                uint32_t LOGADDR;
                uint32_t MSGSADDR;
                struct EEPROMMSGS MSGS;
                /*
                LOGADDR=MAP.addr[10+1];                                           // get the address of the messages pointer
                MSGSADDR=LOGADDR+sizeof(MSGS);                                       // calcukate the messages base address
                eeprom_retrieve(LOGADDR, sizeof(MSGS),&MSGS);                        // reterieve the messages pointers
                sprintf(buffer,"(1)FST %d LST %d",MSGS.fstmsg,MSGS.lstmsg);                          
                bt_send_data(buffer);
                logmsg_add(2,buffer);
                */
 

                bt_send_data("TOP OF DOOR_ACTIVE");


                if(PARMS.USERFID=='Y')
                {

                   
   
                    while(true)
                    {
                    
                        WRONG_TAG=false; 
                        memset(RFIDTAGID,0,sizeof(RFIDTAGID));
                      
                  
                        if(i==0)                                // if there is one
                        {
                              
                            bt_send_data("RFID RETURNED 0");
                            //logmsg_add(2,"RFID RETURNED 0");
                            bt_send_data(RFIDTAGID);
                            //logmsg_add(2,RFIDTAGID);
                            i=strcmp(PARMS.RFTAG,RFIDTAGID);       // check to see if it's the right tag 
                            if(i!=0)                               // if it isn't
                            {
                                WRONG_TAG=true;                    // set the logical
                                logmsg_add(2,"WRONG RFID TAG");
                                DOOR_PHASE=2;
                                break;
                            }
                            DOOR_PHASE=1;
                            break;
                        }
                        else
                            {
                            sprintf(buff,"RFID RETURNED %d",i);
                            //logmsg_add(2,buff);
                            //logmsg_add(2,"RFID RETURNED 0");
                            curr_time=                             // get the realtime clock
                                     dt_to_posix(get_rtc());
                            i=curr_time-DOOR_OPENED_TIME;         // calcuate how long the door was open
                            if(i<121)                             // if it's been less than two minutes 
                            {  
                                DOOR_PHASE=3;                     // set the door phase to 3
                                TIMEOUT=dt_to_posix(get_rtc())+10800; // add three hours to TIMEOUT


                                i=TIMEOUT-dt_to_posix(get_rtc()); // time in seonds until we sound the tag alarm
                                xdate=dt_to_disp(TIMEOUT);                            // testing only - 
                                sprintf(buff,"XXXXXXX %02d:%02d:%02d %02d/%02d/%04d",
                                         xdate.hour,xdate.minute,xdate.second,xdate.month,xdate.day,xdate.year);
                                bt_send_data(buff);

                                if(i<60)                          // if there are less than sity seconds to go
                                    i=i+300;                      //      add five minutes
                                i=i*1000;                         // converted to milliseconds
                                k_timer_start(&tag_alarm,         // set the timer to go off when timeout occurs
                                       K_MSEC(i),K_NO_WAIT);                 
                            }
                            else  DOOR_PHASE=4;
                            break;
                            }

                      

                    }
                            
                }
                sprintf(buff,"DOOR ACTIVE(%d)",DOOR_PHASE);
                bt_send_data(buff);
                //logmsg_add(2,buff);
            if(DOOR_PHASE==4)                                   // if we got here rfid has been enabled and the 
                                                                //    box has been open for more than two minutes   
            {      
                while(true)                                         // so we start trying to get the agent to put the damn thing back
                {
                    no_key_cnt=0;                                   // we ask that the key be replaced
                    while(no_key_cnt<3)                             // three times
                    {

                        no_key_cnt++;    
                        k_timer_stop(&msclock);                     // now ... stop the millisecond timer (it doesn't matter that it's not running)
                        MILLITIME=0;                                // set it to 0
                        k_timer_start(&msclock,                     //  and start it up 
                                  K_MSEC(1),K_MSEC(1));
                        while(true)                                 // now - loop until either 2 seonds have passed or the door has been opened
                        {
                            k_sleep(K_MSEC(50));                    // look twenty times a second
                            i=gpio_pin_get(gpio0,DOORSW);           // see if the door sw is open
                            if(i==0)                                // if it is
                                break;                               //     get out of this loop
                            if(MILLITIME>5000)                      // if five seconds have passed
                            break;                               //     get out of ths loop             
                        }
                        k_timer_stop(&msclock);                     // stop the millisecond clock
                        if(i==0)                                    // if the door was opened
                        {              
                            TIMEOUT=dt_to_posix(get_rtc())+10800;    // we reset the door alarm to three hours from now
                            i=TIMEOUT-dt_to_posix(get_rtc());       // time in seonds until we sound the door alarm
                            i=i*1000;                               // convwerted to milliseconds
                            k_timer_start(&open_alarm,              // set the OPEN DOOR timer
                                K_MSEC(120000),K_NO_WAIT);          // this delay will need to be fixed  
                            k_timer_start(&check_sws,               // restart the check switches routine
                                K_MSEC(500), K_MSEC(500)); 
                            bt_send_data("DOOR OPEN");
                            i=0;                                    // sinc we used i we net to put it back to 0
                            break;                                  // and get out  -  the entire loop starts again         
                        }   
                                  
                    }
                    sprintf(buff,"GOT HERE i=%d",i);
                    bt_send_data(buff);
                    break;
                }
                sprintf(buff,"GOT HERE 2 i=%d",i);
                bt_send_data(buff);
                if(i==0)                                            // if the door was opened during the "please relace the key" loop
                {   
                    vdd2_down();                                     // vdd2 comes doen
                    continue;                                        //      and we go all the way back up to the topof the main loop
                }
                DOOR_PHASE=3;                                        // the door is still closed after asking three times -
                if(!LOGGING)                                               
                     logmsg_add(2, "TAG MISSING NOTIFICATION");      // log the Alert
                if(PARMS.USEAWS=='Y')                                // if we're using AWS
                {                                     
                    strcpy(CMD,"ALERTNO TAG IN LOCK,");              // send out an immediate TAG alert
                    que_send_alert();
                }   
                TIMEOUT=dt_to_posix(get_rtc()) + 10800;               // set TIMEOUT to now + 3 hours
                i=TIMEOUT-dt_to_posix(get_rtc());                   // time in seonds until we sound the tag alarm
                xdate=dt_to_disp(TIMEOUT);                          // testing only - 
                sprintf(buff,"XXXXXXX %02d:%02d:%02d %02d/%02d/%04d",
                            xdate.hour,xdate.minute,xdate.second,xdate.month,xdate.day,xdate.year);
                bt_send_data(buff);

                if(i<60)                                            // and set up the tag alarm - if there are less than sity seconds to go
                    i=i+300;                                        //      add five minutes
                i=i*1000;                                           // converted to milliseconds
                k_timer_start(&tag_alarm,                           // set the timer to go off when timeout occurs
                                K_MSEC(i),K_NO_WAIT);                                        
            }

                if(DOOR_PHASE<4)                                    // if we can lock the door 
                                                                    //   a tag not required
                                                                    //   or  a tag is present 
                                                                    //   or a tag isn't present but the door has been open less than two minutes
                {
                   /*
                    LOGADDR=MAP.addr[10+1];                                           // get the address of the messages pointer
                    MSGSADDR=LOGADDR+sizeof(MSGS);                                       // calcukate the messages base address
                    eeprom_retrieve(LOGADDR, sizeof(MSGS),&MSGS);                        // reterieve the messages pointers
                    sprintf(buffer,"(2)FST %d LST %d",MSGS.fstmsg,MSGS.lstmsg);                             
                    bt_send_data(buffer);
                    logmsg_add(2,buffer); 
                    */
                    DOORACTIVE=false;                                // turn off the switch
                    bt_send_data("HERE 1");    
                    DOOR_OPEN=false;                          // and turn off logical
                    memset(&actrec,0,sizeof(actrec));         // clear the activity record
                    actrec.date=dt_to_posix(get_rtc());       // put the date into the activity record
                    actrec.phase=1;                           // this is a close hasp/door record
                    strcpy(actrec.cmd,"XXXXXX");
                    bt_send_data("HERE 2");
                    motor_cw();                               // rotate the mechanism 90 degrees clockwise          
                    bt_send_data("HERE 3"); 
                    check_center();                           // and go check that the motor is centered
                    bt_send_data("HERE 4"); 
                    if(PARMS.LOCKSTATE!='C')                  // if it didn't cnter properly
                        actrec.err=1;                         // set the error flag on

                    bt_send_data("HERE 5");   
                    actrec.v1=LOW_VOLTAGE;
                    actrec.v2=IDLEVOLTAGE;
 
                    if(DOOR_PHASE==2)                         // if it's the wrong tag
                        actrec.err=11;                        //     it's an error 11
                    if(DOOR_PHASE==3)                         // if there was no rfid tag but the dor closed quickly
                        actrec.err=12;                        //      it's an error 12
                    strcpy(CMD,"ANCHORS,");
                    LOGADDR=MAP.addr[10+1];                                           // get the address of the messages pointer
                    MSGSADDR=LOGADDR+sizeof(MSGS);                                       // calcukate the messages base address
                    eeprom_retrieve(LOGADDR, sizeof(MSGS),&MSGS);                        // reterieve the messages pointers
                    sprintf(buffer,"(3)FST %d LST %d",MSGS.fstmsg,MSGS.lstmsg);                         
                    bt_send_data(buffer);
                    //logmsg_add(2,buffer); 
                    if(!LOGGING)
                        logmsg_add(1, &actrec);                 // go store the activity record
                    LOGADDR=MAP.addr[10+1];                                           // get the address of the messages pointer
                    MSGSADDR=LOGADDR+sizeof(MSGS);                                       // calcukate the messages base address
                    eeprom_retrieve(LOGADDR, sizeof(MSGS),&MSGS);                        // reterieve the messages pointers
                    sprintf(buffer,"(4)FST %d LST %d",MSGS.fstmsg,MSGS.lstmsg);
                    //logmsg_add(2,buffer); 
                    if(PARMS.USEAWS=='Y')                       // if we're using AWS
                    {
                        FLASHER_CONTROL=0;
                        FLASHER_MODE=3;                         // set the flasher mode to flash th GREEN LED
                        k_timer_start(&update_delay,            //       schedule the update
                            K_MSEC(60000),K_MSEC(60000)); 
                        k_timer_start(&flasher,                 // now - we start to slowly flash the LED
                            K_MSEC(200), K_MSEC(200));     
                    }

                if(!CONNECTED)
                    vdd2_down();                                // power down VDD2
                DOORCLOSED=true;
                if(RUNNINGSCRIPT)                               // if we're running a script
                    SCRIPTCMDDONE=true;                         //     let the script task know we're done
                advertising_start();                            // we can start the advertising
                setup_keyscan();                                //     and enable the keypad
                continue;                                       // now .. go back to the top of the main wait loop 
            }


        }
                if(HASPACTIVE)                                   // the hasp switch toggled
                {

                    bt_send_data("HASP ACTIVE");
                    HASPACTIVE=false;
                    k_timer_stop(&check_sws);                    // stop the delay timer (it doesn;t matter if it was never started)
                    k_timer_stop(&spoof);                        // stop the spoof timer (it doesnt matter if it was neveer started)
                    k_timer_stop(&open_alarm);                   // stop the door open timer
                    DOOR_OPEN=false;                             // and turn off logical
                    vdd2_up();                                   // bring up VDD2 (it doesn't matter if it's already up)
                    memset(&actrec,0,sizeof(actrec));            // clear the activity record to NULLS
                    actrec.date=dt_to_posix(get_rtc());          // put the date into the activity record
                    actrec.phase=1;                              // this is a close hasp/door record
                    actrec.v1=LOW_VOLTAGE;                       // put in the voltages
                    actrec.v2=IDLEVOLTAGE;

                    strcpy(actrec.cmd,"XXXXXX");    
                    motor_ccw();                                 // rotate the mechanism 90 degrees counter-clockwise
                    check_center();                              // and go check that the  motor is centered
                    if(PARMS.LOCKSTATE!='C')                     // if it didn't cnter properly
                       actrec.err=1;                             // set the error flag on
                    if(!LOGGING)
                        logmsg_add(1, &actrec);                  // go store the activity record

                    gpio_pin_set(gpio0,LED_GREEN,0);             // turn off the greenled

                    if(PARMS.USEAWS=='Y')                        // if we're using AWS
                    {
                        FLASHER_CONTROL=0;
                        FLASHER_MODE=3;                          // set the flasher mode to flash th GREEN LED
                        k_timer_start(&update_delay,             //       schedue the update
                              K_MSEC(60000),K_MSEC(60000)); 
                        k_timer_start(&flasher,                  // now - we start to slowly flash the LED
                                 K_MSEC(200), K_MSEC(200));     
                        }                  
                    if(!CONNECTED)
                         vdd2_down();                            // power down VDD2
                    if(RUNNINGSCRIPT)                            // if we're running a script
                             SCRIPTCMDDONE=true;                 //     let the script task know we're done
                    advertising_start();                         // so we can start the advertising
                    setup_keyscan();;                             // and enable the keypad

                }
                if(UPDATEPENDING)                                // if there is an uodate pending
                {
                    if(CONNECTED)                                // if bluetooth is connected
                    {
                         bt_send_data("UPDATE PENDING");          //    notify user
                         continue;                                // and loop bak to the top of main                                                    
                    }
                    if(VDD2_UP)                                   // if VDD is up something is going on
                        continue;                                 //     so - loop back to the top of main
                
                    advertising_stop();
                    keypad_disable();
                    k_timer_stop(&update_delay);                   // kill the update timer 
                    k_timer_stop(&flasher);                        // and the flasher                           
                    vdd2_up();                                     // bring up vdd2
                    gpio_pin_set(gpio0,LED_ILLUM,0);               // turn off the illumination LEs
                    que_update();                                  // go post activity to the cloud
                    vdd2_down();                                   // bring vdd2 diwn
                    UPDATEPENDING=false;                           // turn off the update pending fla
                    advertising_start();                         // so we can start the advertising
                    setup_keyscan();                              // and enable the keypad
                    
                }          
	        }
}



void ble_write_thread(void)
{
    uint32_t err;
    struct ble_data_t ble_data, *buf;
    	/* Don't go any further until BLE is initialized */
	k_sem_take(&ble_init_ok, K_FOREVER);
    buf=&ble_data;
   /*
    
	for (;;) {
		// Wait indefinitely for data to be sent over bluetooth 
		buf = k_fifo_get(&fifo_ble_data, K_FOREVER);

		err=bt_nus_send(NULL, buf->data, buf->len);
       // if(err!=0)
	//	 logmsg(2,"BLE WRTE ERR");
     
		
	}
    */
    
    for(;;)
		{
            buf = k_fifo_get(&fifo_ble_data, K_FOREVER);                    // wait for a packet from main
            for(;;)
            {
		         err=bt_nus_send(NULL, buf->data, buf->len);                 //      sent it out to the conected lockbox
		         if(err==-120)                                               // if we got an error 120 (OPERATION IN PROGRESS)
		        {
				  // send_reply("BT BUSY,");
				   k_sleep(K_MSEC(80));
				   continue;
			    }
			break;
            }
             k_free(buf);
		}
       
}

K_THREAD_DEFINE(ble_write_thread_id, STACKSIZE, ble_write_thread, NULL, NULL,
		NULL, PRIORITY, 0, 0);